#define F_CPU 16000000

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include "font6x8.h"


#define RST 2
#define CE 3
#define DC 4
#define DIN 5
#define CLK 6

#define UP 0
#define RIGHT 1
#define DOWN 2
#define LEFT 3

#define MENU 0
#define PLAY 1
#define GAME_OVER 2

#define NORMAL 0
#define LABIRINT 1

#define MAX_LEN 100
#define INITIAL_LEN 20

typedef struct pozitie {
	unsigned int x;
	unsigned int y;
}pozitie;


pozitie snake_body[MAX_LEN];
pozitie head;
pozitie obstacole[MAX_LEN];
pozitie food; //am nevoie doar de colt stanga sus pt food
pozitie bonus_food;
unsigned int scor = 0;
unsigned int snake_direction = RIGHT;
unsigned int current_len = INITIAL_LEN;
unsigned int seed = 0;
unsigned int frame_no = 0;
unsigned int game_state = MENU;
unsigned int game_level = NORMAL;


void LCD_write_command(unsigned char instruction)
{
	unsigned int i;
	PORTD &=~(1 << DC);
	
	
	for (i = 0; i < 8; i++)
	{
		if (instruction & 0x80)
			PORTD |= (1 << DIN);
		else
			PORTD &=~(1 << DIN);
		
		instruction = instruction << 1;
		PORTD &=~(1 << CLK);
		PORTD |= (1 << CLK);
	}
}

void LCD_write_data(unsigned char instruction)
{
	unsigned int i;
	PORTD |=(1 << DC);
	
	
	for (i = 0; i < 8; i++)
	{
		if (instruction & 0x80)
			PORTD |= (1 << DIN);
		else
			PORTD &=~(1 << DIN);
		
		instruction = instruction << 1;
		PORTD &=~(1 << CLK);
		PORTD |= (1 << CLK);
	}
}

int check_food_overlap(pozitie body_position, pozitie food_position, unsigned int type)
{
	if (body_position.x == food_position.x && body_position.y == food_position.y)
		return 1;
	if (body_position.x == food_position.x + 1 && body_position.y == food_position.y + 1)
		return 1;
	
	//daca mancare normala
	if (type == 0) {
		if (body_position.x == food_position.x + 1 && body_position.y == food_position.y)
			return 1;
		if (body_position.x == food_position.x && body_position.y == food_position.y + 1)
			return 1;
	}

	return 0;
}


int check_head_overlap_body()
{
	unsigned int i;
	for ( i = 0; i < current_len; i++)
		if (snake_body[i].x == head.x && snake_body[i].y == head.y)
			return 1;
	return 0;
}

void spawn_food(unsigned int food_type)
{
	seed ++;
	srand(seed);
	pozitie aux_food;
	if (food_type == 0) {
		aux_food = food;
	} else {
		aux_food = bonus_food;
	}
	aux_food.x = rand() % 47;
	aux_food.y = rand() % 83;
	
	if (game_level == LABIRINT) {
		while (aux_food.x  < 2 || aux_food.x > 44 || aux_food.x == 23)  {
			seed ++;
			aux_food.x = rand() % 47;
		}
		while (aux_food.y  < 2 || aux_food.y > 80)  {
			seed ++;
			aux_food.y = rand() % 83;
		}
		
	}
	if (food_type == 0)
		food = aux_food;
	else
		bonus_food = aux_food;
	unsigned int i;
	unsigned int collision = 0;
	
	for (i = 0; i < current_len; i++) {
		
		if (food_type == 0) {
			if (check_food_overlap(snake_body[i], food, food_type) == 1) {
				seed ++;
				collision = 1;
				break;
			}
			if (check_food_overlap(bonus_food, food, food_type) == 1) {
				seed ++;
				collision = 1;
				break;
			}
		}
		if (food_type == 1) {
			if (check_food_overlap(snake_body[i], bonus_food, food_type) == 1) {
				seed ++;
				collision = 1;
				break;
			}
			if (check_food_overlap(food, bonus_food, food_type) == 1) {
				seed ++;
				collision = 1;
				break;
			}
		}
	}
	
	if (collision)
		spawn_food(food_type);
}



void LCD_init(void)
{
	DDRD |=0xFF;   // setez tot DDRd pe 1, adica iesiri
	
	//RST trebuie trecut de la low la high atunci cand este pornit
	//resetez toti registrii pt ca initial sunt undefined
	PORTD &=~(1 << RST);
	_delay_us(1);
	PORTD |= (1 << RST);

	
	PORTD &=~(1 << CE);
	
	
	LCD_write_command(0x21); //LCD extended commands mode
	LCD_write_command(0xB8); // set LCD contrast
	LCD_write_command(0x04); //set temperature coefficient
	LCD_write_command(0x14); //bias mode
	LCD_write_command(0x20); //LCD basic command mode
	LCD_write_command(0x0C);
	
	//head.x = 24;
	//head.y = 42;
	head.x = 19;
	head.y = 42;
	
	
	for (int i = INITIAL_LEN; i > 0; i--) {
		snake_body[i].x = head.x;
		snake_body[i].y = head.y  + i - INITIAL_LEN;
	
	}
	
	spawn_food(0);
}

void follow_head() {
	unsigned int i;
	
	for (i = 0; i < current_len; i++) {
		if (i == current_len - 1) {
			snake_body[i].x = head.x;
			snake_body[i].y = head.y;
		}
		else {
			snake_body[i].x = snake_body[i + 1].x;
			snake_body[i].y = snake_body[i + 1].y;
		}
	}
}

void game_over_msg(char *s)
{
	int i;
	unsigned char c ;
	while (*s) 
	{
		c = *s;
		for ( i = 0; i < 6; i++)
			LCD_write_data(font6x8[c - 32][i]);
		s++;
    }
	
	
}

unsigned int draw_food(pozitie food_position, unsigned int food_type, unsigned int i)
{
	unsigned int segment = 0x00;
	unsigned int mask = 0x01;
	
	if ( i == (food_position.x / 8 * 84 + food_position.y)) {
		mask = 0x01;
		segment |= (mask << (food_position.x % 8));
		mask = 0x01;
		if (food_type == 0 && ((food_position.x + 1) % 8 != 0))
			segment |= (mask << ((food_position.x + 1) % 8));
	}
	
	if ( i == (food_position.x / 8 * 84 + (food_position.y + 1))) {
		mask = 0x01;
		if (food_type == 0)
			segment |= (mask << (food_position.x % 8));
		mask = 0x01;
		if ((food_position.x + 1) % 8 != 0)
			segment |= (mask << ((food_position.x + 1) % 8));
	}

	return segment;
}

unsigned int draw_border(unsigned int i)
{
	unsigned int segment = 0x00;
	
	if (i % 84 == 1 || i % 84 == 82)
		segment |=  0xFF;
	
	if (i < 82 && i > 1)
		segment |= 0x01; 
	
	if (i < 502 && i > 421)
		segment |= 0x80;
	return segment;
	
}

int check_collision_with_labyrinth()
{
	if ((head.x == 0 || head.x == 47 )&& (head.y >= 1 && head.y <= 82))
		return 1;
	if ((head.y == 1 || head.y == 82) && (head.x >= 1 && head.x <= 46))
		return 1;
	if (head.x == 23 && (head.y < 40  || head.y >= 42))
		return 1;
	return 0;
}

void LCD_clear(void)          // clear the LCD
{
	unsigned int i;
			

	for (i = 0; i < 504; i++)
	{
	   LCD_write_data(0x00);
	}		
}
int main(void) {
	
	LCD_init();
	
	//set as input pins
	DDRB &= ~(1 << PB0);
	DDRB &= ~(1 << PB1);
	DDRB &= ~(1 << PB2);
	DDRB &= ~(1 << PB3);
	unsigned int segment;
	unsigned int mask, wait_bonus = 0;
	unsigned int i, j,flag = 3, bonus_active = 0;
	char s[5];
	//LCD_clear();
	while(1) {

		if (game_state == MENU) {
			if (game_level == NORMAL) {
				game_over_msg(" SELECT LEVEL ");
				game_over_msg(" *NORMAL      ");
				game_over_msg("  LABIRINT      ");
			}
			else {
				game_over_msg(" SELECT LEVEL ");
				game_over_msg("  NORMAL      ");
				game_over_msg(" *LABIRINT      ");
			}
			 _delay_ms(400);
			for (i = 0; i < 504 - (44 * 6); i++)
				LCD_write_data(0x00);
			
			if ((PINB & (1 << PB3)) == 0) {
				//snake_direction = DOWN;
				game_level ^= 1;
				
			}
			if ((PINB & (1 << PB1)) == 0) {
				//snake_direction = UP;
				game_level ^= 1;
			}
			if ((PINB & (1 << PB2)) == 0) {
				game_state = PLAY;
			}
			
		}
		if (game_state == PLAY) {
			
			if ((PINB & (1 << PB0)) == 0) {
				seed ++;
				if (snake_direction != RIGHT)
					snake_direction = LEFT;
			}
			if ((PINB & (1 << PB1)) == 0) {
				seed ++;
				if (snake_direction != DOWN)
					snake_direction = UP;
			}
			if ((PINB & (1 << PB2)) == 0) {
				seed ++;
				if (snake_direction != LEFT)
					snake_direction = RIGHT;
			}
			if ((PINB & (1 << PB3)) == 0) {
				seed ++;
				if (snake_direction != UP)
					snake_direction = DOWN;
			}
			if (scor != 0 && bonus_active == 0 && wait_bonus == 0) {
				spawn_food(1);
				bonus_active = 1;
			}
			flag = 3;
			if (check_food_overlap(head, food, 0) == 1)
				flag = 0;
			if (bonus_active && check_food_overlap(head, bonus_food, 1) == 1) {
				flag = 1;
				bonus_active = 0;
				wait_bonus = 0;
			}
			if (flag != 3) {
				snake_body[current_len].x = head.x;
				snake_body[current_len].y = head.y;
				
				current_len += 1;
				if (flag == 0) {
					scor ++;
					spawn_food(0);
				}
				else {
					scor += 10;
					spawn_food(1);
				}
				
			}
			else
				follow_head();
			if (snake_direction == RIGHT) {
				if (head.y != 83)
					head.y += 1;
				else
					head.y = 0;
			}
			else if (snake_direction == DOWN) {
				if (head.x != 47)
					head.x += 1;
				else
					head.x = 0;
			}
			else if (snake_direction == UP) {
				if (head.x != 0)
					head.x -= 1;
				else
					head.x = 47;
			}
			else if (snake_direction == LEFT) {
				if (head.y != 0)
					head.y -= 1;
				else
					head.y = 83;
			}
			
			
			if ((check_head_overlap_body() == 1) == 1) {
				game_state = GAME_OVER;
			}
			
			if (game_level == LABIRINT && (check_collision_with_labyrinth() == 1)) {
				game_state = GAME_OVER;
			}
			
			if (game_state == PLAY) {
				for(i = 0; i < 504; i++) {
					segment = 0x00;
					mask = 0x01;
					if (i == (head.x / 8 * 84 + head.y)) {
						segment = mask << (head.x % 8);
					}
					
					for (j = 0; j < current_len; j++) {
						mask = 0x01;
						if (i == (snake_body[j].x / 8 * 84 + snake_body[j].y)) {
							segment |= (mask << (snake_body[j].x % 8));
						}
					}
					
					segment |= draw_food(food, 0, i);
					
					if (bonus_active == 1)
						segment |= draw_food(bonus_food, 1, i);
					
					if (game_level == LABIRINT) {
						segment |= draw_border(i);
						
						if (i >= (2 * 84 + 2)  && i <= (2 * 84 + 2 + 37))
							segment |= 0x80;
						if (i > (2 * 84 + 2 + 39)  && i <= (2 * 84 + 2 + 84 - 4))
							segment |= 0x80;
					}
					
					LCD_write_data(segment);
				}
			
				if (bonus_active == 1) {
					frame_no ++;
					if (frame_no == 75) {
						frame_no = 0;
						bonus_active = 0;
						wait_bonus = 1;
					}
				}
			
				else {
					wait_bonus ++;
				}
				if (wait_bonus == 100) {
					wait_bonus = 0;
					
				}
				_delay_ms(80);
			}
		
		}
		if (game_state == GAME_OVER) {
			game_over_msg(" GAME OVER    ");
			game_over_msg(" SCOR: ");
			
			dtostrf(scor, 3, 0, s);
			game_over_msg(s);
			
			for (i = 0; i < 504 - (24 * 6); i++)
				LCD_write_data(0x00);
			_delay_ms(80);
			
			if ((PINB & (1 << PB2)) == 0) {
				game_state = MENU;
				current_len = 	INITIAL_LEN;
				game_level = NORMAL;
				seed = 0;
				scor = 0;
				
				wait_bonus = 0;
				bonus_active = 0;
				flag = 3;
				LCD_init();
				
				
			}
		}
		
	}
	
	
}